"""Main PAC methods.

This file include the following methods :
- Mean Vector Length (Canolty, 2006)
- Kullback Leibler Distance (Tort, 2010)
- Heights Ratio (Lakatos, 2005)
- Normalized direct Pac (Ozkurt, 2012)
- Phase Synchrony (Penny, 2008; Cohen, 2008)
"""

import numpy as np
from scipy.special import erfinv

__all__ = ['ComputePac']


def ComputePac(pha, amp, idp, nbins, p, optimize):
    """Copute real Phase-Amplitude coupling.

    Each method take at least a pha and amp array with the respective
    dimensions:
    pha.shape = (npha, ..., npts)
    amp.shape = (namp, ..., npts)
    And each method should return a (namp, npha, ...)
    """
    if idp == 1:  # Mean Vector Length (Canolty, 2006)
        return MVL(pha, amp, optimize)

    elif idp == 2:  # Kullback-Leiber distance (Tort, 2010)
        return klDistance(pha, amp, nbins, optimize)

    elif idp == 3:  # Heights ratio (Lakatos, 2005)
        return HeightsRatio(pha, amp, nbins, optimize)

    elif idp == 4:  # ndPac (Ozkurt, 2012)
        return ndPac(pha, amp, p, optimize)

    elif idp == 5:  # Phase-Synchrony (Penny, 2008; Cohen, 2008)
        return PhaseSync(pha, amp, optimize)

    else:
        raise ValueError(str(idp) + " is not recognized as a valid pac "
                         "method.")


def MVL(pha, amp, optimize):
    """Mean Vector Length (Canolty, 2006).

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

    Return:
        pac: np.ndarray
            PAC of shape (npha, namp, ...)
    """
    # Number of time points :
    npts = pha.shape[-1]
    return np.abs(np.einsum('i...j, k...j->ik...', amp, np.exp(1j*pha),
                            optimize=optimize)) / npts


def klDistance(pha, amp, nbins, optimize):
    """Kullback Leibler Distance (Tort, 2010).

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

        nbins: int
            Number of bins in which the phase in cut in bins.

    Return:
        pac: np.ndarray
            PAC of shape (npha, namp, ...)
    """
    # Get the phase locked binarized amplitude :
    p_j = _kl_hr(pha, amp, nbins, optimize)
    # Divide the binned amplitude by the mean over the bins :
    p_j /= p_j.sum(axis=0, keepdims=True)
    # Take the log of non-zero values :
    p_j = p_j * np.ma.log10(p_j).filled(-np.inf)
    # Compute the PAC :
    pac = 1 + p_j.sum(axis=0)/np.log10(nbins)
    # Set distribution distances that are really closed to zero :
    pac[np.isinf(pac)] = 0.
    return pac


def HeightsRatio(pha, amp, nbins, optimize):
    """Pac heights ratio (Lakatos, 2005).

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

        nbins: int
            Number of bins in which the phase in cut in bins.

    Return:
        pac: np.ndarray
            PAC of shape (npha, namp, ...)
    """
    # Get the phase locked binarized amplitude :
    p_j = _kl_hr(pha, amp, nbins, optimize)
    # Divide the binned amplitude by the mean over the bins :
    p_j /= p_j.sum(axis=0, keepdims=True)
    # Find (maxximum, minimum) of the binned distribution :
    h_max, h_min = p_j.max(axis=0), p_j.min(axis=0)
    # Compute pac :
    pac = (h_max - h_min) / h_max
    return pac


def _kl_hr(pha, amp, nbins, optimize):
    """Binarize the amplitude according to phase values.

    This function is shared by the Kullback-Leibler Distance and the
    Height Ratio.
    """
    # Build the default binned phase vector :
    binsize = (2 * np.pi) / float(nbins)
    vecbin = np.arange(-np.pi, np.pi, binsize)

    abin = []
    for i in vecbin:
        # Find where phase take vecbin values :
        idx = np.logical_and(pha >= i, pha < i + binsize)
        # Take the sum of amplitude inside the bin :
        abin.append(np.einsum('i...j, k...j->ik...', amp, idx,
                              optimize=optimize))

    return np.array(abin)


def ndPac(pha, amp, p, optimize):
    """Normalized direct Pac (Ozkurt, 2012).

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

        p: float
            The p-value to use.

    Return:
        pac: np.ndarray
            PAC of shape (npha, namp, ...)
    """
    npts = amp.shape[-1]
    # Normalize amplitude :
    np.subtract(amp, np.mean(amp, axis=-1, keepdims=True), out=amp)
    np.divide(amp, np.std(amp, axis=-1, keepdims=True), out=amp)
    # Compute pac :
    pac = np.abs(np.einsum('i...j, k...j->ik...', amp, np.exp(1j*pha),
                           optimize=optimize))
    pac *= pac/npts
    # Set to zero non-significant values:
    xlim = erfinv(1-p)**2
    pac[pac <= 2 * xlim] = 0.
    return pac


def PhaseSync(pha, amp, optimize):
    """Phase Synchrony (Penny, 2008; Cohen, 2008).

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

    Return:
        pac: np.ndarray
            PAC of shape (npha, namp, ...)
    """
    # Number of time points :
    npts = pha.shape[-1]
    pac = np.einsum('i...j, k...j->ik...', np.exp(-1j*amp), np.exp(1j*pha),
                    optimize=optimize)
    return np.abs(pac)/npts



"""This file contains several functions for memory usage.

Taken from the numpy tricks : http://ipython-books.github.io/featured-01/
"""
import numpy as np


__all__ = ['id', 'arrays_share_data']


def id(x):
    """Returns the memory block address of an array."""
    return x.__array_interface__['data'][0]


def get_data_base(arr):
    """For a given Np array, finds the base array that "owns" the actual data.
    """
    base = arr
    while isinstance(base.base, np.ndarray):
        base = base.base
    return base


def arrays_share_data(x, y):
    """Return if two arrays share an offset."""
    return get_data_base(x) is get_data_base(y)



"""Simply get the name of defined methods."""

__all__ = ['pacstr']


def pacstr(idpac):
    """Return correspond methods string."""
    # Pac methods :
    if idpac[0] == 1:
        method = 'Mean Vector Length (MVL, Canolty, 2006)'
    elif idpac[0] == 2:
        method = 'Kullback-Leiber Distance (KLD, Tort, 2010)'
    elif idpac[0] == 3:
        method = 'Heights ratio (HR, Lakatos, 2005)'
    elif idpac[0] == 4:
        method = 'ndPac (Ozkürt, 2012)'
    elif idpac[0] == 5:
        method = 'Phase-Synchrony (Cohen, 2008; Penny, 2008)'
    else:
        raise ValueError("No corresponding pac method.")

    # Surrogate method :
    if idpac[1] == 0:
        suro = 'No surrogates'
    elif idpac[1] == 1:
        suro = 'Swap phase/amplitude across trials'
    elif idpac[1] == 2:
        suro = 'Swap amplitude blocks across time'
    elif idpac[1] == 3:
        suro = 'Shuffle amplitude time-series'
    elif idpac[1] == 4:
        suro = 'Time lag'
    else:
        raise ValueError("No corresponding surrogate method.")

    # Normalization methods :
    if idpac[2] == 0:
        norm = 'No normalization'
    elif idpac[2] == 1:
        norm = 'Substract the mean of surrogates'
    elif idpac[2] == 2:
        norm = 'Divide by the mean of surrogates'
    elif idpac[2] == 3:
        norm = 'Substract then divide by the mean of surrogates'
    elif idpac[2] == 4:
        norm = "Substract the mean and divide by the deviation of the " + \
               "surrogates"
    else:
        raise ValueError("No corresponding normalization method.")

    return method, suro, norm



"""
PHASE-AMPLITUDE TOOLBOX
=======================

Tensorpac is an open-source Python toolbox designed for computing
Phase-Amplitude Coupling. 
"""
from .pac import Pac
from .utils import PacSignals, PacVec, PacTriVec

__all__ = ['Pac', 'PacSignals', 'PacVec', 'PacTriVec']



"""Normalize PAC by surrogates methods.

This file include the following methods :
- No normalization
- Substraction : substract the mean of surrogates
- Divide : divide by the mean of surrogates
- Substract then divide : substract then divide by the mean of surrogates
- Z-score : substract the mean and divide by the deviation of the
            surrogates
"""

__all__ = ['normalize']


def normalize(pac, sMean, sStd, idn):
    """List of the normalization methods.

    Use a normalization to normalize the true cfc value by the surrogates.
    Here's the list of the normalization methods :
    - No normalization
    - Substraction : substract the mean of surrogates
    - Divide : divide by the mean of surrogates
    - Substract then divide : substract then divide by the mean of surrogates
    - Z-score : substract the mean and divide by the deviation of the
                surrogates

    The normalized method only return the normalized cfc.
    """
    if idn == 0:  # No normalisation
        return pac

    elif idn == 1:  # Substraction
        return pac-sMean

    elif idn == 2:  # Divide
        return pac/sMean

    elif idn == 3:  # Substract then divide
        pac -= sMean
        pac /= sMean
        return pac
        # return (pac-sMean)/sMean

    elif idn == 4:  # Z-score
        return (pac-sMean)/sStd



"""Test tensorpac utils."""
import numpy as np
from tensorpac.utils import PacVec, PacSignals, PacTriVec


def test_pacvec():
    """Definition of PAC vectors."""
    assert PacVec()
    assert PacVec(fpha=(1, 30, 2, 2), famp=(60, 200, 10, 5))
    assert PacVec(fpha=[1, 2], famp=np.arange(50))
    assert PacVec(fpha=np.array([[2, 4], [5, 7], [9, 10]]),
                  famp=np.array([[30, 60], [60, 90], [100, 200]]).T)
    assert PacVec(fpha=[[1, 2], [5, 7]], famp=[60, 150])


def test_pacsignals_dTrials():
    """Definition of artificially coupled signals using dPha/dAmp."""
    assert PacSignals(fpha=5, famp=130, sf=512, ndatasets=23, chi=0.9, noise=2,
                      dpha=35, damp=46)


def test_pacsignals_bandwidth():
    """Definition of artificially coupled signals using bandwidth."""
    assert PacSignals(fpha=[5, 7], famp=[30, 60], sf=200., ndatasets=100,
                      chi=0.5, noise=3., npts=1000)


def test_defaultArgs():
    """Test default aurguments for PacVec."""
    assert PacSignals(chi=2., noise=11., dpha=120., damp=200.)


def test_trivec():
    """Definition of triangular vectors."""
    assert PacTriVec(2, 200, 10)



"""Utility PAC functions.

- PacSignals : generate artificially phase-amplitude coupled signals
- PacVec : generate cross-frequency coupling vectors
"""
import numpy as np

__all__ = ['PacSignals', 'PacVec', 'PacTriVec']


###############################################################################
###############################################################################
#                             SIGNALS
###############################################################################
###############################################################################

def PacSignals(fpha=10, famp=100, sf=1024, npts=4000, ndatasets=10, chi=0,
               noise=1, dpha=0, damp=0):
    """Generate artificially phase-amplitude coupled signals.

    Kargs:
        fpha: int/float, optional, [def: 10]
            Frequency for phase

        famp: int/float, optional, [def: 100]
            Frequency for amplitude

        sf: int, optional, [def: 1024]
            Sampling frequency

        ndatasets : int, optional, [def: 10]
            Number of datasets

        npts: int, optional, [def: 4000]
            Number of points for each signal.

        chi: int/float (0<=chi<=1), optional, [def: 0]
            Amount of coupling. If chi=0, signals of phase and amplitude
            are strongly coupled.

        noise: int/float (0<=noise<=3), optional, [def: 1]
            Amount of noise

        dpha: int/float (0<=dpha<=100), optional, [def: 0]
            Introduce a random incertitude on the phase frequency.
            If fpha is 2, and dpha is 50, the frequency for the phase signal
            will be between :
            [2-0.5*2, 2+0.5*2]=[1,3]

        damp: int/float (0<=damp<=100), optional, [def: 0]
            Introduce a random incertitude on the amplitude frequency.
            If famp is 60, and damp is 10, the frequency for the amplitude
            signal will be between :
            [60-0.1*60, 60+0.1*60]=[54,66]

    Return:
        data: array
            The randomly coupled signals. The shape of data will be
            (ndatasets x npts)

        time: array
            The corresponding time vector
    """
    # Check the inputs variables :
    if not 0 <= chi <= 1:
        chi = 0
    if not 0 <= noise <= 3:
        noise = 0
    if not 0 <= dpha <= 100:
        dpha = 0
    if not 0 <= damp <= 100:
        damp = 0
    fpha, famp = np.asarray(fpha), np.asarray(famp)
    time = np.mgrid[0:ndatasets, 0:npts][1] / sf
    data = np.zeros_like(time)

    # Band / Delta parameters :
    if fpha.ndim == 0:
        aPha = [fpha*(1-dpha/100), fpha*(1+dpha/100)]
        deltaPha = aPha[0] + (aPha[1]-aPha[0])*np.random.rand(ndatasets, 1)
    elif fpha.ndim == 1:
        deltaPha = np.random.uniform(fpha[0], fpha[1], ndatasets)
    if famp.ndim == 0:
        aAmp = [famp*(1-damp/100), famp*(1+damp/100)]
        deltaAmp = aAmp[0] + (aAmp[1]-aAmp[0])*np.random.rand(ndatasets, 1)
    elif famp.ndim == 1:
        deltaAmp = np.random.uniform(famp[0], famp[1], ndatasets)

    # Reshape phase/amplitude bands :
    deltaPha, deltaAmp = deltaPha.reshape(-1, 1), deltaAmp.reshape(-1, 1)

    # Create phase and amplitude signals :
    xl = np.sin(2 * np.pi * deltaPha * time)
    xh = np.sin(2 * np.pi * deltaAmp * time)

    # Create the coupling :
    ah = .5 * ((1. - chi) * xl + 1. + chi)
    al = 1.

    # Generate datasets :
    data = (ah * xh) + (al * xl)
    data += noise * np.random.rand(*data.shape)  # Add noise

    return data, time[0, :]

###############################################################################
###############################################################################
#                             FREQUENCY VECTOR
###############################################################################
###############################################################################


def PacVec(fpha=(2, 30, 2, 1), famp=(60, 200, 10, 5)):
    """Generate cross-frequency coupling vectors.

    Kargs:
        fpha: tuple, optional, [def: (2, 30, 2, 1)]
            Frequency parameters for phase. Each argument inside the tuple
            mean (starting fcy, ending fcy, width, step)

        famp: tuple, optional, [def: (60, 200, 10, 5)]
            Frequency parameters for amplitude. Each argument inside the
            tuple mean (starting fcy, ending fcy, width, step)

    Returns:
        pVec: np.ndarray, shape (N, 2)
            Array containing the pairs of phase frequencies.

        aVec: np.ndarray, shape (N, 2)
            Array containing the pairs of amplitude frequencies.
    """
    return _CheckFreq(fpha), _CheckFreq(famp)


def _CheckFreq(f):
    """Check the frequency definition."""
    f = np.atleast_2d(np.asarray(f))
    #
    if len(f.reshape(-1)) == 1:
        raise ValueError("The length of f should at least be 2.")
    elif 2 in f.shape:  # f of shape (N, 2) or (2, N)
        if f.shape[1] is not 2:
            f = f.T
    elif np.squeeze(f).shape == (4,):  # (fstart, fend, fwidth, fstep)
        f = _CreatePairsVector(*tuple(np.squeeze(f)))
    else:  # Sequential
        f = f.reshape(-1)
        f.sort()
        f = np.c_[f[0:-1], f[1::]]
    return f


def _CreatePairsVector(fstart, fend, fwidth, fstep):
    # Generate two array for phase and amplitude :
    fdown = np.arange(fstart, fend-fwidth, fstep)
    fup = np.arange(fstart+fwidth, fend, fstep)
    return np.c_[fdown, fup]


def PacTriVec(fstart=60, fend=160, fwidth=10):
    """Generate triangular vector.

    Kargs:
        fstart: float, optional, (def: 60)
            Starting frequency.

        fend: float, optional, (def: 160)
            Ending frequency.

        fwidth: float, optional, (def: 10)
            Frequency bandwidth.

    Returns:
        f: np.ndarray
            The triangular vector.

        tridx: np.ndarray
            The triangular index for the reconstruction.
    """
    starting = np.arange(fstart, fend+fwidth, fwidth)
    f, tridx = np.array([]), np.array([])
    for num, k in enumerate(starting[0:-1]):
        # Lentgh of the vector to build :
        L = len(starting) - (num + 1)
        # Create the frequency vector for this starting frequency :
        fst = np.c_[np.full(L, k), starting[num+1::]]
        nfst = fst.shape[0]
        # Create the triangular index for this vector of frequencies :
        idx = np.c_[np.flipud(np.arange(nfst)), np.full(nfst, num)]
        tridx = np.concatenate((tridx, idx), axis=0) if tridx.size else idx
        f = np.concatenate((f, fst), axis=0) if f.size else fst
    return f, tridx



"""Statistic tools."""
import numpy as np
from scipy.stats import chi2


def pearson(x, y, st='i...j, k...j->ik...', optimize=True):
    """Pearson correlation for multi-dimensional arrays.

    Args:
        x, y: np.ndarray
            Compute pearson correlation between the multi-dimensional arrays
            x and y.

    Kargs:
        st: string, optional, (def: 'i..j, k..j->ik...')
            The string to pass to the np.einsum function.

        optimize: bool, optional, (def: True)
            Optimize argument of the np.einsum function. Use either False,
            True, 'greedy' or 'optimal'.

    Returns:
        cov: np.ndarray
            The pearson correlation array.
    """
    n = x.shape[-1]
    # Distribution center :
    mu_x = x.mean(-1, keepdims=True)
    mu_y = y.mean(-1, keepdims=True)
    # Distribution deviation :
    s_x = x.std(-1, ddof=n-1, keepdims=True)
    s_y = y.std(-1, ddof=n-1, keepdims=True)
    # Compute correlation coefficient :
    cov = np.einsum(st, x, y, optimize=optimize)
    mu_xy = np.einsum(st, mu_x, mu_y, optimize=optimize)
    cov -= n * mu_xy
    cov /= np.einsum(st, s_x, s_y, optimize=optimize)
    return cov


def circ_corrcc(alpha, x, optimize=True):
    """Correlation coefficient between a circular and a linear random variable.

    Code from the Circular Statistics Toolbox for Matlab By Philipp Berens 2009
    and adapted for multi-dimensional arrays.

    Args:
        alpha: vector
            Sample of angles in radians

        x: vector
            Sample of linear random variable

        optimize: bool, optional, (def: True)
            Optimize argument of the np.einsum function. Use either False,
            True, 'greedy' or 'optimal'.

    Returns:
        rho: float
            Correlation coefficient.

        pval: float
            P-value.
    """
    n = alpha.shape[-1]
    # Compute correlation coefficient for sin and cos independently
    sa, ca = np.sin(alpha), np.cos(alpha)
    rxs = pearson(x, sa, optimize=optimize)
    rxc = pearson(x, ca, optimize=optimize)
    rcs = pearson(sa, ca, st='i...j, k...j->i...', optimize=optimize)
    rcs = rcs[np.newaxis, ...]

    # Compute angular-linear correlation (equ. 27.47)
    rho = np.sqrt((rxc**2 + rxs**2 - 2*rxc*rxs*rcs)/(1-rcs**2))

    # Compute pvalue :
    pval = 1. - chi2.cdf(n*rho**2, 2)

    return rho, pval



"""Test tensorpac functions."""
import numpy as np
from tensorpac import Pac
from tensorpac.utils import PacTriVec


def test_IdpacDefinition():
    """Test Pac object definition."""
    nmeth, nsuro, nnorm = 5, 5, 5
    for k in range(nmeth):
        for i in range(nsuro):
            for j in range(nnorm):
                p = Pac(idpac=(k+1, i, j))
                str(p)


def test_filteringDefinition():
    """Test filtering defintion."""
    dcomplex = ['hilbert', 'wavelet']
    filt = ['butter', 'fir1', 'bessel']
    cycle = (12, 24)
    filtorder = 4
    width = 12
    for k in dcomplex:
        for i in filt:
            Pac(dcomplex=k, filt=i, filtorder=filtorder, cycle=cycle,
                width=width)


def test_spectral():
    """Test filtering using the provided filters."""
    data = np.random.rand(3, 1000)
    p = Pac()
    dcomplex = ['hilbert', 'wavelet']
    filt = ['butter', 'fir1', 'bessel']
    for k in dcomplex:
        for i in filt:
            p.dcomplex = k
            p.filt = i
            p.filter(1024, data, axis=1, njobs=1)


def test_pacMeth():
    """Test all Pac methods."""
    pha = np.random.rand(2, 7, 1024)
    amp = np.random.rand(3, 7, 1024)
    nmeth, nsuro, nnorm = 5, 5, 5
    p = Pac()
    for k in range(nmeth):
        for i in range(nsuro):
            for j in range(nnorm):
                p.idpac = (k+1, i, j)
                p.fit(pha, amp, axis=2, traxis=1, njobs=1, nperm=2)


def test_compute():
    """Test filtering test computing PAC."""
    data = np.random.rand(2, 1024)
    p = Pac(idpac=(4, 0, 0))
    p.filterfit(1024, data, data, njobs=1)
    p.idpac = (1, 1, 1)
    p.filterfit(1024, data, data, njobs=1, nperm=2, correct=True)
    p.dcomplex = 'wavelet'
    p.filter(1024, data, axis=1, njobs=1)


def test_properties():
    """Test Pac properties."""
    p = Pac()
    # Idpac :
    p.idpac
    p.idpac = (2, 1, 1)
    # Filt :
    p.filt
    p.filt = 'butter'
    # Dcomplex :
    p.dcomplex
    p.dcomplex = 'wavelet'
    # Cycle :
    p.cycle
    p.cycle = (12, 24)
    # Filtorder :
    p.filtorder
    p.filtorder = 6
    # Width :
    p.width
    p.width = 12


def test_pacComodulogram():
    """Test Pac object definition.

    This test works locally but failed on travis...
    """
    try:
        f, tridx = PacTriVec()
        pac = np.random.rand(20, 10)
        pval = np.random.rand(20, 10)
        p = Pac(fpha=np.arange(11), famp=np.arange(21))
        print(len(p.xvec), len(p.yvec))
        p.comodulogram(pac, rmaxis=True, dpaxis=True)
        p.comodulogram(pac, plotas='contour', pvalues=pval)
        p.comodulogram(pac, plotas='pcolor', pvalues=pval, levels=[.5, .7],
                       under='gray', over='red', bad='orange')
        p = Pac(fpha=np.arange(11), famp=f)
        p.polar(pac, np.arange(10), np.arange(20), interp=.8)
        pac = np.random.rand(len(f))
        p.triplot(pac, f, tridx)
        # p.savefig('file.png', dpi=10)
        # p.show()
    except:
        pass


def test_PreferredPhase():
    """Test the prefered phase."""
    data = np.random.rand(2, 1024)
    p = Pac(idpac=(4, 0, 0))
    pha = p.filter(1024, data, axis=1, ftype='phase')
    amp = p.filter(1024, data, axis=1, ftype='amplitude')
    p.pp(pha, amp, axis=2)


def test_ERPAC():
    """Test the ERPAC."""
    data = np.random.rand(2, 1024)
    p = Pac(idpac=(4, 0, 0))
    pha = p.filter(1024, data, axis=1, ftype='phase')
    amp = p.filter(1024, data, axis=1, ftype='amplitude')
    p.erpac(pha, amp, traxis=1)



"""Main surrogates estimation methods.

This file include the following methods :
- No surrogates
- Swap phase/amplitude across trials
- Swap amplitude blocks across time
- Shuffle amplitude and phase time-series
- Shuffle phase time-series
- Shuffle amplitude time-series
- Time lag
"""

import numpy as np
from joblib import Parallel, delayed
from .methods import ComputePac

__all__ = ['ComputeSurogates']


def ComputeSurogates(pha, amp, surargs, pacargs, nperm, njobs):
    """Compute surrogates using tensors and parallel computing.

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

        suragrs: tuple
            Tuple containing the arguments to pass to the suroSwitch function.

        pacargs: tuple
            Tuple containing the arguments to pass to the ComputePac function.

        nperm: int
            Number of permutations.

        njobs: int
            Number of jos for the parallel computing.

    Returns:
        suro: np.ndarray
            Array of pac surrogates of shape (nperm, npha, namp, ..., npts)
    """
    s = Parallel(n_jobs=njobs)(delayed(_computeSur)(
                            pha, amp, surargs, pacargs) for k in range(nperm))
    return np.array(s)


def _computeSur(pha, amp, surargs, pacargs):
    """Compute surrogates.

    This is clearly not the optimal implementation. Indeed, for each loop the
    suroSwicth and ComputePac have a several "if" that slow down the execution,
    at least a little bit. And, it's not esthetic but joblib doesn't accept
    to pickle functions.

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

        suragrs: tuple
            Tuple containing the arguments to pass to the suroSwitch function.

        pacargs: tuple
            Tuple containing the arguments to pass to the ComputePac function.
    """
    # Get the surrogates :
    pha, amp = suroSwitch(pha, amp, *surargs)
    # Compute PAC on surrogates :
    return ComputePac(pha, amp, *pacargs)


def suroSwitch(pha, amp, idn, axis, traxis, nblocks):
    """List of methods to compute surrogates.

    The surrogates are used to normalized the cfc value. It help to determine
    if the cfc is reliable or not. Usually, the surrogates used the same cfc
    method on surrogates data.
    Here's the list of methods to compute surrogates:
    - No surrogates
    - Swap phase/amplitude across trials
    - Swap amplitude blocks across time.
    - Shuffle amplitude time-series
    - Time lag
    """
    # No surrogates
    if idn == 0:
        return None

    # Swap phase/amplitude across trials :
    elif idn == 1:
        return SwapPhaAmp(pha, amp, traxis)

    # Swap amplitude :
    elif idn == 2:
        return SwapBlocks(pha, amp, axis, nblocks)

    # Shuffle amplitude values
    elif idn == 3:
        return ShuffleAmp(pha, amp, axis)

    # Introduce a time lag
    elif idn == 4:
        return TimeLag(pha, amp, axis)

    else:
        raise ValueError(str(idn) + " is not recognized as a valid surrogates"
                         " evaluation method.")


###############################################################################
###############################################################################
#                            SWAPING
###############################################################################
###############################################################################


def SwapPhaAmp(pha, amp, axis):
    """Swap phase/amplitude trials (Tort, 2010).

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

        axis: int
            Location of the trial axis.

    Return:
        pha: np.ndarray
            Swapped version of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Swapped version of amplitudes of shapes (namp, ..., npts)
    """
    return _dimswap(pha, axis), _dimswap(amp, axis)


def SwapBlocks(pha, amp, axis, nblocks):
    """Swap amplitudes time blocks.

    To reproduce (Bahramisharif, 2013), use a number of blocks of 2.

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

        axis: int
            Location of the time axis.

        nblocks: int
            Number of blocks to in which the amplitude is splitted.

    Return:
        pha: np.ndarray
            Original version of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Swapped version of amplitudes of shapes (namp, ..., npts)
    """
    # Split amplitude across time into two parts :
    ampl = np.array_split(amp, nblocks, axis=axis)
    # Revered elements :
    ampl.reverse()
    return pha, np.concatenate(ampl, axis=axis)


###############################################################################
###############################################################################
#                            SHUFFLING
###############################################################################
###############################################################################


def ShuffleAmp(pha, amp, axis):
    """Randomly shuffle amplitudes across time.

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

        axis: int
            Location of the time axis.

    Return:
        pha: np.ndarray
            Original version of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Shuffled version of amplitudes of shapes (namp, ..., npts)
    """
    return pha, _dimswap(amp, axis)


def TimeLag(pha, amp, axis):
    """Introduce a time lag on phase series..

    Args:
        pha: np.ndarray
            Array of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Array of amplitudes of shapes (namp, ..., npts)

        axis: int
            Location of the time axis.

    Return:
        pha: np.ndarray
            Shiffted version of phases of shapes (npha, ..., npts)

        amp: np.ndarra
            Original version of amplitudes of shapes (namp, ..., npts)
    """
    npts = pha.shape[-1]
    return np.roll(pha, np.random.randint(npts), axis=axis), amp


def _dimswap(x, axis=0):
    """Swap values into an array at a specific axis.

    Args:
        x: np.ndarray
            Array of data to swap

    Kargs:
        axis: int, optional, (def: 0)
            Axis along which to perform swapping.

    Returns:
        x: np.ndarray
            Swapped version of x.
    """
    # Dimension vector :
    dimvec = [slice(None)] * x.ndim
    # Random integer vector :
    rndvec = np.arange(x.shape[axis])
    np.random.shuffle(rndvec)
    dimvec[axis] = rndvec
    # Return a swapped version of x :
    return x[dimvec]



"""Main PAC class."""
import numpy as np
from scipy.signal import hilbert

from .utils import PacVec
from .pacstr import pacstr
from .spectral import spectral
from .methods import ComputePac, _kl_hr
from .surrogates import ComputeSurogates
from .normalize import normalize
from .visu import PacPlot
from .stats import circ_corrcc


class Pac(PacPlot):
    """Compute Phase-Amplitude Coupling (PAC) using tensors.

    Computing PAC is assessed in three steps : compute the real PAC, compute
    surrogates and finally, because PAC is very sensible to the noise, correct
    the real PAC by the surrogates. This implementation is modular i.e. it lets
    you choose among a large range of possible combinations.

    Kargs:
        idpac: tuple/list, optional, (def: (1, 1, 3))
            Choose the combination of methods to use in order to extract PAC.
            This tuple must be composed of three integers where each one them
            refer

            * First digit: refer to the pac method:

                - '1': Mean Vector Length (MVL) [#f1]_
                - '2': Kullback-Leibler Distance (KLD) [#f2]_
                - '3': Heights Ratio (HR) [#f3]_
                - '4': ndPAC [#f4]_
                - '5': Phase Synchrony [#f3]_

            * Second digit: refer to the method for computing surrogates:

                - '0': No surrogates
                - '1': Swap phase/amplitude across trials [#f2]_
                - '2': Swap amplitude time blocks [#f5]_
                - '3': Shuffle amplitude time-series
                - '4': Time lag [#f1]_

            * Third digit: refer to the normalization method for correction:

                - '0': No normalization
                - '1': Substract the mean of surrogates
                - '2': Divide by the mean of surrogates
                - '3': Substract then divide by the mean of surrogates
                - '4': Z-score

        fpha, famp: list/tuple/array, optional, (def: [2, 4] and [60, 200])
            Frequency vector for the phase and amplitude. Here you can use
            several forms to define those vectors :

                * Basic list/tuple (ex: [2, 4] or [8, 12]...)
                * List of frequency bands (ex: [[2, 4], [5, 7]]...)
                * Dynamic definition : (start, stop, width, step)
                * Range definition (ex : np.arange(3) => [[0, 1], [1, 2]])

        dcomplex: string, optional, (def: 'hilbert')
            Method for the complex definition. Use either 'hilbert' or
            'wavelet'.

        filt: string, optional, (def: 'fir1')
            Filtering method (only if dcomplex is 'hilbert'). Choose either
            'fir1', 'butter' or 'bessel'

        cycle: tuple, optional, (def: (3, 6))
            Control the number of cycles for filtering (only if dcomplex is
            'hilbert'). Should be a tuple of integers where the first one
            refers to the number of cycles for the phase and the second for the
            amplitude [#f5]_.

        filtorder: int, optional, (def: 3)
            Filter order for the Butterworth and Bessel filters (only if
            dcomplex is 'hilbert').

        width: int, optional, (def: 7)
            Width of the Morlet's wavelet.

        nbins: int, optional, (def: 18)
            Number of bins for the KLD and HR PAC method [#f2]_ [#f3]_

        nblocks: int, optional, (def: 2)
            Number of blocks for splitting the amplitude. Only active is
            the surrogate method is 2 [#f5]_.

    .. warning::
        * The ndPac [#f4]_ include a fast and reliable statistical test. As a
          result, if the ndPAC is choosed as the main PAC method, surrogates
          and normalization will be deactivate.

        * The phase in a particular frequency band can either be extracted
          using wavelet convolution or filtering followed by the Hilbert
          transform. As a result, every filtering related input (cycle, filt,
          filtorder) are going to be active if the complex decomposition is
          Hilbert.

    Methods:
        self.filt:
            Filt the data in the specified frequency bands.

        self.fit:
            Run the PAC on filtered data.

        self.filtfit:
            Filt the data then compute PAC on it.

        self.comodulogram:
            Plot PAC.

    .. rubric:: Footnotes
    .. [#f1] `Canolty et al, 2006 <http://www.ncbi.nlm.nih.gov/pmc/articles/
       PMC2628289/>`_
    .. [#f2] `Tort et al, 2010 <http://www.ncbi.nlm.nih.gov/pmc/articles/
       PMC2941206/>`_
    .. [#f3] `Lakatos et al, 2005 <https://www.ncbi.nlm.nih.gov/pubmed/
       15901760>`_
    .. [#f4] `Ozkurt et al, 2012 <http://www.ncbi.nlm.nih.gov/pubmed/
       22531738/>`_
    .. [#f5] `Bahramisharif et al, 2013 <http://www.jneurosci.org/content/33/
       48/18849.short/>`_

    """

    ###########################################################################
    #                              __FCN__
    ###########################################################################
    def __init__(self, idpac=(1, 1, 3), fpha=[2, 4], famp=[60, 200],
                 dcomplex='hilbert', filt='fir1', cycle=(3, 6), filtorder=3,
                 width=7, nbins=18, nblocks=2):
        """Check and initialize."""
        # ----------------- CHECKING -----------------
        # Pac methods :
        self._idcheck(idpac)
        # Frequency checking :
        self.fpha, self.famp = PacVec(fpha, famp)
        self.xvec, self.yvec = self.fpha.mean(1), self.famp.mean(1)

        # Check spectral properties :
        self._speccheck(filt, dcomplex, filtorder, cycle, width)

        # ----------------- SELF -----------------
        self.nbins, self.nblocks = int(nbins), int(nblocks)

    def __str__(self):
        """String representation."""
        st = self.method
        st += '\n' + self.surro if self.surro else ''
        st += '\n' + self.norm if self.norm else ''
        return st

    ###########################################################################
    #                              METHODS
    ###########################################################################
    def filter(self, sf, x, axis=-1, ftype='phase', keepfilt=False, njobs=-1):
        """Filt the data in the specified frequency bands.

        Args:
            sf: float
                The sampling frequency.

            x: np.ndarray
                Array of data.

        Kargs:
            axis: int, optional, (def: -1)
                Location of the time axis.

            ftype: string, optional, (def: 'phase')
                Specify if you want to extract phase ('phase') or the amplitude
                ('amplitude').

            njobs: int, optional, (def: -1)
                Number of jobs to compute PAC in parallel. For very large data,
                set this parameter to 1 in order to prevent large memory usage.

        keepfilt: bool, optional, (def: False)
            Specify if you only want the filtered data (True). This parameter
            is only avaible with dcomplex='hilbert' and not wavelet.

        Returns:
            xfilt: np.ndarray
                The filtered data of shape (n_frequency, ...)
        """
        # Sampling frequency :
        if not isinstance(sf, (int, float)):
            raise ValueError("The sampling frequency must be a float number.")
        else:
            sf = float(sf)
        # Compatibility between keepfilt and wavelet :
        if (keepfilt is True) and (self._dcomplex is 'wavelet'):
            raise ValueError("Using wavelet for the complex decomposition do "
                             "not allow to get filtered data only. Set the "
                             "keepfilt parameter to False or set dcomplex to "
                             "'hilbert'.")
        # Switch between phase or amplitude :
        if ftype is 'phase':
            tosend = 'pha' if not keepfilt else None
            xfilt = spectral(x, sf, self.fpha, axis, tosend, self._dcomplex,
                             self._filt, self._filtorder, self._cycle[0],
                             self._width, njobs)
        elif ftype is 'amplitude':
            tosend = 'amp' if not keepfilt else None
            xfilt = spectral(x, sf, self.famp, axis, tosend, self._dcomplex,
                             self._filt, self._filtorder, self._cycle[1],
                             self._width, njobs)
        else:
            raise ValueError("ftype must either be None, 'phase' or "
                             "'amplitude.'")
        return xfilt

    def fit(self, pha, amp, axis=1, traxis=0, nperm=200, optimize=True,
            get_surro=False, correct=False, njobs=-1):
        """Compute PAC on filtered data.

        Args:
            pha, amp: np.ndarray
                Array of filtered data with respectively a shape of (npha, ...)
                and (namp, ...). If you want to compute PAC locally i.e. on the
                same electrode, x=pha=amp. For distant coupling, pha and
                amp could be different but still must to have the same shape.

        Kargs:
            axis: int, optional, (def: 1)
                Dimension where is located the time axis. By default, the axis
                will be consider as well.

            traxis: int, optional, (def: 0)
                Dimension where is located the trial axis. By default the next-
                to-last axis is consider as the trial axis.

            nperm: int, optional, (def: 200)
                Number of surrogates to compute.

            optimize: bool, optional, (def: True)
                Optimize argument of the np.einsum function. Use either False,
                True, 'greedy' or 'optimal'.

            get_surro: bool, optional, (def: False)
                Return surrogate chance distribution.

            correct: bool, optional, (def: True)
                Correct the PAC estimation XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

            njobs: int, optional, (def: -1)
                Number of jobs to compute PAC in parallel. For very large data,
                set this parameter to 1 in order to prevent large memory usage.

        Returns:
            pac: np.ndarray
                Phase-Amplitude Coupling measure of shape (namp, npha, ...).

            pvalue: np.ndarray
                P-values (None if no surrogates)

            suro: np.ndarray
                If get_suro is True, get the chance distribution of shape
                (nperm, namp, npha, ...)

        .. warning::
            * Surrogates are only going to be computed if the second and third
              digits are no 0.

            * The ndPAC use a p value and every non-significant PAC estimation
              is set to zero. This p value is computed as 1/nperm.

            * The traxis argument is only used if you picked up the surrogates
              method 1: "swap phase and amplitude trials [#f2]_"

            * Basically, the surrogate evaluation proposed by [#f5]_ split the
              amplitude into two equal parts, then swap those two blocks. But
              the nblocks parameter allow to split into a larger number.
        """
        # Check phase and amplitude :
        pha, amp, axis = self._phampcheck(pha, amp, axis)
        # For the phase synchrony, extract the phase of the amplitude :
        if self._idpac[0] == 5:
            amp = np.angle(hilbert(amp, axis=axis))
        suro, pvalues = None, None
        # Compute pac :
        pacargs = (self.idpac[0], self.nbins, 1/nperm, optimize)
        pac = ComputePac(pha, amp, *pacargs)

        # Compute surrogates (if needed) :
        if self._csuro:
            surargs = (self.idpac[1], axis, traxis, self.nblocks)
            suro = ComputeSurogates(pha, amp, surargs, pacargs, nperm, njobs)

            # Get the mean / deviation of surrogates :
            m_surro, std_surro = np.mean(suro, axis=0), np.std(suro, axis=0)

            # Normalize pac by surrogates :
            pac = normalize(pac, m_surro, std_surro, self.idpac[2])

            # Compute statistics :
            suro.sort(0)
            suro -= pac[np.newaxis, ...]
            pvalues = 1 - np.sum(suro < 0, axis=0)/nperm
            pvalues[pvalues < 1/nperm] = 1/nperm

        if self._idpac[0] == 4:
            pvalues = np.ones_like(pac)
            pvalues[np.nonzero(pac)] = 1/nperm

        if correct:
            pac[pac < 0.] = 0.

        if get_surro:
            return pac, pvalues, suro
        else:
            return pac, pvalues

    def filterfit(self, sf, xpha, xamp, axis=1, traxis=0, nperm=200,
                  optimize=True, get_surro=False, correct=False, njobs=-1):
        """Filt the data then compute PAC on it.

        Args:
            sf: float
                The sampling frequency.

            xpha, xamp: np.ndarray
                Array of data for computing PAC. xpha is the data used for
                extracting phases and xamp, amplitudes. Both arrays must have
                the same shapes. If you want to compute PAC locally i.e. on the
                same electrode, x=xpha=xamp. For distant coupling, xpha and
                xamp could be different but still must to have the same shape.

        Kargs:
            axis: int, optional, (def: 1)
                Dimension where is located the time axis. By default, the axis
                will be consider as well.

            traxis: int, optional, (def: 0)
                Dimension where is located the trial axis. By default the next-
                to-last axis is consider as the trial axis.

            nperm: int, optional, (def: 200)
                Number of surrogates to compute.

            optimize: bool, optional, (def: True)
                Optimize argument of the np.einsum function. Use either False,
                True, 'greedy' or 'optimal'.

            get_surro: bool, optional, (def: False)
                Return surrogate chance distribution.

            correct: bool, optional, (def: True)
                Correct the PAC estimation XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

            njobs: int, optional, (def: -1)
                Number of jobs to compute PAC in parallel. For very large data,
                set this parameter to 1 in order to prevent large memory usage.

        Returns:
            pac: np.ndarray
                Phase-Amplitude Coupling measure of shape (namp, npha, ...).

            pvalue: np.ndarray
                P-values (None if no surrogates)

            suro: np.ndarray
                If get_suro is True, get the chance distribution of shape
                (nperm, namp, npha, ...)

        .. warning::
            * Surrogates are only going to be computed if the second and third
              digits are no 0.

            * The ndPAC use a p value and every non-significant PAC estimation
              is set to zero. This p value is computed as 1/nperm.

            * The traxis argument is only used if you picked up the surrogates
              method 1: "swap phase and amplitude trials [#f2]_"

            * Basically, the surrogate evaluation proposed by [#f5]_ split the
              amplitude into two equal parts, then swap those two blocks. But
              the nblocks parameter allow to split into a larger number.
        """
        # Shape checking :
        if xpha.shape != xamp.shape:
            raise ValueError("The shape of xpha and xamp must be equals.")
        # Extract phase (npha, ...) and amplitude (namp, ...) :
        pha = self.filter(sf, xpha, axis, 'phase', False, njobs)
        amp = self.filter(sf, xamp, axis, 'amplitude', False, njobs)

        # Special cases :
        if self._idpac[0] == 5:
            amp = np.angle(hilbert(amp, axis=-1))

        # Compute pac :
        return self.fit(pha, amp, axis+1, traxis+1, nperm, optimize,
                        get_surro, correct, njobs)

    def pp(self, pha, amp, axis=-1, nbins=72, optimize=True):
        """Compute the preferred-phase.

        Args:
            pha: np.ndarray
                Phase of slower oscillations.

            amp: np.ndarray
                Amplitude of fastest oscillations.

        Kargs:
            axis: int, optional, (def: -1)
                Location of the time axis.

            nbins: int, optional, (def: 72)
                Number of bins for bining the amplitude according to phase
                slices.

            optimize: bool, optional, (def: True)
                Optimize argument of the np.einsum function. Use either False,
                True, 'greedy' or 'optimal'.

        Returns:
            ampbin: np.ndarray
                The binned amplitude according to the phase of shape
                (nbins, namp, npha...).

            pp: np.ndarray
                The prefered phase where the amplitude is maximum of shape
                (namp, npha, ...).

            polarvec: np.ndarray
                The phase vector for the polar plot.
        """
        # Check phase and amplitude shapes :
        pha, amp, axis = self._phampcheck(pha, amp, axis)
        # Define the method name :
        self.method, self.surro, self.norm = 'Preferred-Phase (PP)', '', ''
        # Move the time axis to the end :
        pha = np.moveaxis(pha, axis, -1)
        amp = np.moveaxis(amp, axis, -1)
        # Bin the amplitude according to the phase :
        ampbin = _kl_hr(pha, amp, nbins, optimize)
        ampbin /= ampbin.sum(axis=0, keepdims=True)
        # Find the index where the amplitude is maximum over the bins :
        idxmax = ampbin.argmax(axis=0)
        # Find the preferred phase :
        binsize = (2 * np.pi) / float(nbins)
        vecbin = np.arange(-np.pi, np.pi, binsize) + binsize/2
        pp = vecbin[idxmax]
        # Build the phase vector (polar plot) :
        polarvec = np.linspace(-np.pi, np.pi, ampbin.shape[0])
        return ampbin, pp, polarvec

    def erpac(self, pha, amp, traxis=0, optimize=True):
        """Compute the Event-Related Phase-Amplitude Coupling (ERPAC).

        The ERPAC [#f6]_ is used to measure PAC across trials and is
        interesting for real-time estimation.

        Args:
            pha: np.ndarray
                Phase of slower oscillations.

            amp: np.ndarray
                Amplitude of fastest oscillations.

        Kargs:
            traxis: int, optional, (def: 0)
                Location of the trial axis.

            optimize: bool, optional, (def: True)
                Optimize argument of the np.einsum function. Use either False,
                True, 'greedy' or 'optimal'.

        Returns:
            erpac: np.ndarray
                The ERPAC estimation.

            pvalue: np.ndarray
                The associated p-values.

        .. warning::
            ERPAC is computed across trials, therefor, the it does not use an
            *axis* variable but instead, a *traxis* variable which specify
            where is located the axis to consider as trials.

        .. [#f6] `Voytek et al, 2013 <https://www.ncbi.nlm.nih.gov/pubmed/
           22986076>`_
        """
        # Check the phase/amplitude :
        if (pha.ndim <= 1) or (amp.ndim <= 1):
            raise ValueError("The phase and amplitude must have at least two"
                             " dimensions (trials, time).")
        pha, amp, _ = self._phampcheck(pha, amp, traxis)
        # Get method name :
        self.method = "Event-Related Phase-Amplitude Coupling (ERPAC, " + \
                      "Voytek et al. 2013)"
        self.surro, self.norm = '', ''
        # Move the trial axis to the end :
        pha = np.swapaxes(pha, traxis, -1)
        amp = np.swapaxes(amp, traxis, -1)
        # Compute the correlation between the circular phase and le linear
        # amplitude :
        return circ_corrcc(pha, amp, optimize=optimize)

    ###########################################################################
    #                              CHECKING
    ###########################################################################
    def _idcheck(self, idpac):
        """Check the idpac parameter."""
        idpac = np.atleast_1d(idpac)
        self._csuro = True
        if not all([isinstance(k, int) for k in idpac]) and (len(idpac) != 3):
            raise ValueError("idpac must be a tuple/list of 3 integers.")
        else:
            # Ozkurt PAC case :
            if idpac[0] == 4:
                idpac = np.array([4, 0, 0])
                self._csuro = False
            if (idpac[1] == 0) or (idpac[2] == 0):
                self._csuro = False
                idpac = (idpac[0], 0, 0)
        self._idpac = idpac
        self.method, self.surro, self.norm = pacstr(idpac)

    def _speccheck(self, filt=None, dcomplex=None, filtorder=None, cycle=None,
                   width=None):
        """Check spectral parameters."""
        # Check the filter name :
        if filt is not None:
            if filt not in ['fir1', 'butter', 'bessel']:
                raise ValueError("filt must either be 'fir1', 'butter' or "
                                 "'bessel'")
            else:
                self._filt = filt
        # Check cycle :
        if cycle is not None:
            cycle = np.asarray(cycle)
            if (len(cycle) is not 2) or not cycle.dtype == int:
                raise ValueError("Cycle must be a tuple of two integers.")
            else:
                self._cycle = cycle
        # Check complex decomposition :
        if dcomplex is not None:
            if dcomplex not in ['hilbert', 'wavelet']:
                raise ValueError("dcomplex must either be 'hilbert' or "
                                 "'wavelet'.")
            else:
                self._dcomplex = dcomplex
        # Convert filtorder :
        if filtorder is not None:
            self._filtorder = int(filtorder)
        # Convert Morlet's width :
        if width is not None:
            self._width = int(width)

    def _phampcheck(self, pha, amp, axis):
        """Check phase and amplitude values."""
        # Shape checking :
        if pha.ndim != amp.ndim:
            raise ValueError("pha and amp must have the same number of "
                             "dimensions.")
        # Force phase / amplitude to be at least (1, N) :
        if (pha.ndim == 1) and (amp.ndim == 1):
            pha = pha.reshape(1, -1)
            amp = amp.reshape(1, -1)
            axis = 1
        # Check if the phase is in radians :
        if np.ptp(pha) > 2 * np.pi:
            raise ValueError("Your phase is probably in degrees and should be"
                             " converted in radians using either np.degrees or"
                             " np.deg2rad.")
        # Check if the phase/amplitude have the same number of points on axis:
        if pha.shape[axis] != amp.shape[axis]:
            phan, ampn = pha.shape[axis], amp.shape[axis]
            raise ValueError("The phase ("+str(phan)+") and the amplitude "
                             "("+str(ampn)+") do not have the same number of "
                             "points on the specified axis ("+str(axis)+").")
        # Force the phase to be in [-pi, pi] :
        pha = (pha + np.pi) % (2 * np.pi) - np.pi
        return pha, amp, axis

    ###########################################################################
    #                              PROPERTIES
    ###########################################################################
    # ----------- IDPAC -----------
    @property
    def idpac(self):
        """Get the idpac value."""
        return self._idpac

    @idpac.setter
    def idpac(self, value):
        """Set idpac value."""
        self._idcheck(value)

    # ----------- FILT -----------
    @property
    def filt(self):
        """Get the filt value."""
        return self._filt

    @filt.setter
    def filt(self, value):
        """Set filt value."""
        self._speccheck(filt=value)

    # ----------- DCOMPLEX -----------
    @property
    def dcomplex(self):
        """Get the dcomplex value."""
        return self._dcomplex

    @dcomplex.setter
    def dcomplex(self, value):
        """Set dcomplex value."""
        self._speccheck(dcomplex=value)

    # ----------- CYCLE -----------
    @property
    def cycle(self):
        """Get the cycle value."""
        return self._cycle

    @cycle.setter
    def cycle(self, value):
        """Set cycle value."""
        self._speccheck(cycle=value)

    # ----------- FILTORDER -----------
    @property
    def filtorder(self):
        """Get the filtorder value."""
        return self._filtorder

    @filtorder.setter
    def filtorder(self, value):
        """Set filtorder value."""
        self._speccheck(filtorder=value)

    # ----------- WIDTH -----------
    @property
    def width(self):
        """Get the width value."""
        return self._width

    @width.setter
    def width(self, value):
        """Set width value."""
        self._width = value



"""Extract spectral informations from data."""
import numpy as np
from joblib import Parallel, delayed
from scipy.signal import hilbert
from .filtering import filtdata

__all__ = ['spectral']


def spectral(x, sf, f, axis, stype, dcomplex, filt, filtorder, cycle, width,
             njobs):
    """Extract spectral informations from data.

    Args:
        x: np.ndarray
            Array of data

        sf: float
            Sampling frequency

        f: np.ndarray
            Frequency vector of shape (N, 2)

        axis: int
            Axis where the time is located.

        stype: string
            Spectral informations to extract (use either 'pha' or 'amp')

        dcomplex: string
            Complex decomposition type. Use either 'hilbert' or 'wavelet'

        filt: string
            Name of the filter to use (only if dcomplex is 'hilbert'). Use
            either 'eegfilt', 'butter' or 'bessel'.

        filtorder: int
            Order of the filter (only if dcomplex is 'hilbert')

        cycle: int
            Number of cycles to use for fir1 filtering.

        width: int
            Width of the wavelet.

        njobs: int
            Number of jobs to use. If jobs is -1, all of them are going to be
            used.
    """
    # Filtering + complex decomposition :
    if dcomplex is 'hilbert':
        # Filt each time series :
        nf = range(f.shape[0])
        xf = Parallel(n_jobs=njobs)(delayed(filtdata)(
                  x, sf, f[k, :], axis, filt, cycle, filtorder) for k in nf)
        # Use hilbert for the complex decomposition :
        xd = hilbert(xf, axis=axis+1) if stype is not None else np.array(xf)
    elif dcomplex is 'wavelet':
        f = f.mean(1)  # centered frequencies
        xd = Parallel(n_jobs=njobs)(delayed(morlet)(
                                            x, sf, k, axis, width) for k in f)

    # Extract phase / amplitude :
    if stype is 'pha':
        return np.angle(np.moveaxis(xd, axis+1, -1))
    elif stype is 'amp':
        return np.abs(np.moveaxis(xd, axis+1, -1))
    elif stype is None:
        return np.moveaxis(xd, axis+1, -1)


def morlet(x, sf, f, axis=0, width=7.):
    """Complex decomposition of a signal x using the morlet wavelet.

    Args:
        x: np.ndarray, shape (N,)
            The signal to use for the complex decomposition. Must be
            a vector of length N.

        sf: float
            Sampling frequency

        f: np.ndarray, shape (2,)
            Frequency vector

    Kargs:
        width: float, optional, (def: 7.)
            Width of the wavelet

        axis: int, optional, (def: 0)
            Axis along performing the convolution.

    Returns:
        xout: np.ndarray, shape (N,)
            The complex decomposition of the signal x.
    """
    dt = 1 / sf
    sf = f / width
    st = 1 / (2 * np.pi * sf)

    # Build morlet's wavelet :
    t = np.arange(-width * st / 2, width * st / 2, dt)
    A = 1 / np.sqrt((st * np.sqrt(np.pi)))
    m = A * np.exp(-np.square(t) / (2 * np.square(st))) * np.exp(
                                                       1j * 2 * np.pi * f * t)

    def ndmorlet(xt):
        # Compute morlet :
        y = np.convolve(xt, m)
        return y[int(np.ceil(len(m) / 2)) - 1:int(len(y) - np.floor(
                                                                len(m) / 2))]
    return np.apply_along_axis(ndmorlet, axis, x)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from setuptools import setup, find_packages
import pip
from pip.req import parse_requirements
from optparse import Option

__version__ = "0.5.2"
NAME = 'Tensorpac'
AUTHOR = "Etienne Combrisson"
MAINTAINER = "Etienne Combrisson"
EMAIL = 'e.combrisson@gmail.com'
KEYWORDS = "phase-amplitude-coupling pac tensor"
DESCRIPTION = "Tensor-based Phase-Amplitude Coupling"
URL = 'http://etiennecmb.github.io/tensorpac/'
DOWNLOAD_URL = 'https://github.com/EtienneCmb/tensorpac/archive/v' + __version__ + '.tar.gz'
# Data path :
PACKAGE_DATA = {}


def read(fname):
    """Read README and LICENSE."""
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


options = Option('--workaround')
options.skip_requirements_regex = None
REQ_FILE = './requirements.txt'
# Hack for old pip versions: Versions greater than 1.x
# have a required parameter "sessions" in parse_requierements
if pip.__version__.startswith('1.'):
    install_reqs = parse_requirements(REQ_FILE, options=options)
else:
    from pip.download import PipSession  # pylint:disable=E0611
    options.isolated_mode = False
    install_reqs = parse_requirements(REQ_FILE,  # pylint:disable=E1123
                                      options=options,
                                      session=PipSession)

REQS = [str(ir.req) for ir in install_reqs]

setup(
    name=NAME,
    version=__version__,
    packages=find_packages(),
    package_dir={'tensorpac': 'tensorpac'},
    package_data=PACKAGE_DATA,
    include_package_data=True,
    description=DESCRIPTION,
    long_description=read('README.rst'),
    platforms='any',
    setup_requires=['numpy', 'joblib'],
    install_requires=REQS,
    dependency_links=[],
    author=AUTHOR,
    maintainer=MAINTAINER,
    author_email=EMAIL,
    url=URL,
    download_url=DOWNLOAD_URL,
    license=read('LICENSE'),
    keywords=KEYWORDS,
    classifiers=["Development Status :: 3 - Alpha",
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: Education',
                 'Intended Audience :: Developers',
                 'Topic :: Scientific/Engineering :: Visualization',
                 "Programming Language :: Python :: 3.5"
                 ])
